--[[
		3D lib by RoboLV
		Dev started on 2015.07.13
]]



local M = {}

-- Variables

--== Vector3D lib
M.vec = require("vec3D")
M.objParser = require("obj_loader")
local vec = M.vec

--== Camera
M.cam = {
	pos = vec(0,0,0),
	lookAt = vec(0,0,1), -- direction
	ang = vec()
}
function M.cam:SetPos( vec )
	
	self.pos = vec
	
end


--== Main Mesh
--[[ Struct:

	|-[]					- Mesh object
	   |- []				- Array of vertex
	       |- []			- Vertex
		       |- pos1		- Relative position
			   |- pos2		- Relative position
			   |- pos3		- Relative position
			   |- UV[2]		- UV array x,y
			   |- color		- Vertex color
			   |- material	- Image/ material
			   |- lightmap	- Light map texture
	
]]
M.mesh = {}

--== Save all loaded model ( vertexes )
M.modelsBuffer = {}

--== Models
--[[ Struct:
		|- []			- Model object
]]
M.models = {}

--==Options
M.options = {
	-- 1) render by lines
	-- 2) render by polygons( triangles)
	-- 3) render by dots
	-- 4) rayTracing method ( experimental )
	rendertype = 1,
	height = love.graphics.getHeight(),
	width = love.graphics.getWidth(),
	halfHeight = love.graphics.getHeight() / 2,
	halfWidth = love.graphics.getWidth() / 2,
}

-- Functions

--== Main Initialization function
function M:Initialize()
	
end

--== Main Render function
function M:Render()
	
	local screenFix = vec( self.options.halfWidth, self.options.halfHeight, 0 )
	local scaleFix = vec( self.options.width, self.options.height, 0 )
	local vecDiv3 = vec(3,3,3)
	
	local vec1= vec()
	local vec2= vec()
	local vec3= vec()
	
	local whiteColor = vec(255,255,255)
	
	-- Draw mesh
	
	for _,mesh in pairs( self.mesh ) do
		local model = mesh.model
		
		local viewPos = model:GetPos() + self.cam.pos
		local pos1 = mesh.pos1 + viewPos
		local pos2 = mesh.pos2 + viewPos
		local pos3 = mesh.pos3 + viewPos
		
		-- 3D to 2D space
		vec1:setX( pos1:x() / pos1:z() * scaleFix:x() + screenFix:x() )
		vec1:setY( -pos1:y() / pos1:z() * scaleFix:y() + screenFix:y() )
		vec2:setX( pos2:x() / pos2:z() * scaleFix:x() + screenFix:x() )
		vec2:setY( -pos2:y() / pos2:z() * scaleFix:y() + screenFix:y() )
		vec3:setX( pos3:x() / pos3:z() * scaleFix:x() + screenFix:x() )
		vec3:setY( -pos3:y() / pos3:z() * scaleFix:y() + screenFix:y() )
		
		local color = model.color * ( mesh.color or whiteColor )/whiteColor
		love.graphics.setColor(color:x(), color:y(),color:z(), model.alpha )
		
		if self.options.rendertype == 1 then
			love.graphics.line( math.floor(vec1:x()), math.floor(vec1:y()), math.floor(vec2:x()), math.floor(vec2:y()) )
			love.graphics.line(	math.floor(vec2:x()), math.floor(vec2:y()), math.floor(vec3:x()), math.floor(vec3:y()) ) 
			love.graphics.line( math.floor(vec3:x()), math.floor(vec3:y()), math.floor(vec1:x()), math.floor(vec1:y()) )
		
		elseif self.options.rendertype == 2 then
			love.graphics.polygon( 'fill', 	math.floor(vec1:x()), math.floor(vec1:y()),
											math.floor(vec2:x()), math.floor(vec2:y()),
											math.floor(vec3:x()), math.floor(vec3:y()) )
		
		elseif self.options.rendertype == 3 then
			love.graphics.circle("fill", math.floor(vec1:x()), math.floor(vec1:y()), math.floor(20/pos1:z()), math.floor(20/pos1:z()))
			love.graphics.circle("fill", math.floor(vec2:x()), math.floor(vec2:y()), math.floor(20/pos2:z()), math.floor(20/pos2:z()))
			love.graphics.circle("fill", math.floor(vec3:x()), math.floor(vec3:y()), math.floor(20/pos3:z()), math.floor(20/pos3:z()))
		
		end
		
	end
end

--== Main update function
function M:Update( dt )
	
	-- calculate camera
	
	-- sort
	self:MeshSort()
	
end

function M:PrintTable( tab, deph )
	deph = deph or 0
	local ttyp = type( tab )
	if ttyp == "table" and tab.type ~= "vec" then
		for k,v in pairs( tab ) do
			local out = ""
			for i = 0, deph do out = out .. "  " end
			out = out.."[]"..tostring(k)
			print( out )
			self:PrintTable( v, deph + 1 )
		end
	elseif ttyp ~= "function" then
		local out = ""
		for i = 0, deph do out = out .. "  " end
		out = out.."|-"..tostring(tab)
		print( out )
	end
end

function M:MeshSort()
	
	table.remove( self.mesh )
	self.mesh = {}
	local vecDiv3 = vec(3,3,3)
	
	for _,model in pairs( self.models ) do
		for _,mesh in pairs( model.mesh ) do
			--print(_, mesh.pos1, mesh.pos2, mesh.pos3 )
			if mesh.pos1 and mesh.pos2 and mesh.pos3 then
				
				local xpos = mesh.xmidle + model:GetPos()
				local xlook = ( xpos + self.cam.pos )
				local dot = mesh.normal:dot(xlook)*-1
				
				if dot > 0  then
					mesh.toCam = ( xpos - self.cam.pos ):getLength()
					mesh.xpos = xpos
					mesh.model = model
					--mesh.dot = dot
					
					table.insert( self.mesh, mesh )
				end
			end
		end
	end
	-- Long sort !!!!
	--( 100 000 per ~0.4 sek) http://lua-users.org/wiki/LuaSorting
	table.sort(self.mesh, function(a,b) return a.toCam > b.toCam  end)
	
end

--== Draw 3D line
function M:drawLine( vec1, vec2 )
	local a,b
	local screenFix = vec( self.options.halfWidth, self.options.halfHeight, 0 )
	local scaleFix = vec( self.options.width, self.options.height, 0 )
	
	local a = vec( 	vec1:x() / vec1:z(),
						vec1:y() / vec1:z(),
						0
	)
	local b = vec( 	vec2:x() / vec2:z(),
						vec2:y() / vec2:z(),
						0
	)
	a = a * scaleFix + screenFix
	b = b * scaleFix + screenFix
	
	love.graphics.setColor(255,100,0 )
	love.graphics.line( a:x(), a:y(), b:x(), b:y() )
	
end

--== Create model
function M:newModel()
	local model = vec():deepcopy( self.base.model )
	model.Space = self
	table.insert( self.models, model )
	return model
end

function string:split(sep)
	local sep, fields = sep or ":", {}
	local pattern = string.format("([^%s]+)", sep)
	self:gsub(pattern, function(c) fields[#fields+1] = c end)
	return fields
end

function M:LoadModel( file )
	file = file
	print( file )
	print( love.filesystem.getUserDirectory( ) )
	local content = {}
	if love.filesystem.exists( file ) then
		
		self.modelsBuffer[file] = {}
		
		for line in love.filesystem.lines(file) do
			table.insert( content, line )
		end
		
		local obj = self.objParser.parse( content )
		--self:PrintTable( obj )
		
		for k,v in pairs( obj.f ) do
			
			local vert1 = v[1].v
			local vert2 = v[2].v
			local vert3 = v[3].v
			
			local mesh = {}
			mesh.color = vec( math.random(255), math.random(255), math.random(255) )
			mesh.pos1 = obj.v[ vert1 ]
			mesh.pos2 = obj.v[ vert2 ]
			mesh.pos3 = obj.v[ vert3 ]
			
			if mesh.pos1 and mesh.pos2 and mesh.pos3 then
				
				mesh.xmidle = (mesh.pos1 + mesh.pos2 + mesh.pos3) / vec(3,3,3) 
				mesh.normal =  mesh.pos1:PolyNormal( mesh.pos2, mesh.pos3 )
				
				table.insert( self.modelsBuffer[file], mesh )
			end
		end
		
	end
end

-- Base classes
M.base = {}

--== Model base class
local MODEL = {}
MODEL.pos 	= vec(0,0,0)
MODEL.ang 	= vec(0,0,0)
MODEL.mesh 	= {}
MODEL.scale	= vec(1,1,1)
MODEL.color = vec(255,255,255)
MODEL.alpha = 255

--++ Set model ( vertexes, ect.. from memory ) ("models/player.obj")
function MODEL:SetModel( name )

	if not self.Space.modelsBuffer[name] then return end
	
	self.mesh = self.pos:deepcopy( self.Space.modelsBuffer[name] )
	self.mesh.model = self
end

--++ Set material from loaded
function MODEL:SetMaterial( name )
	
end

--++ Basic function
function MODEL:Initialize() end
function MODEL:GetPos() return vec():deepcopy(self.pos) end
function MODEL:SetPos(vector) self.pos = vec():deepcopy(vector) end
function MODEL:GetAng() return vec():deepcopy(self.ang) end
function MODEL:SetAng(vector) self.ang = vec():deepcopy(vector) end
function MODEL:GetScale() return vec():deepcopy(self.scale) end
function MODEL:SetScale(vector) self.scale = vec():deepcopy(vector) end
function MODEL:GetMesh() return self.ang end
function MODEL:SetMesh( mesh ) self.mesh = vec():deepcopy(mesh) end

M.base.model = MODEL

-- Module end
return M